#    Copyright Reliance Jio Infocomm, Ltd.
#    Author: Soren Hansen <Soren.Hansen@ril.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import argparse
import etcd
import errno
import socket
import sys
import time
import urllib3
from urllib3.exceptions import HTTPError

class DeploymentOrchestrator(object):
    def __init__(self, host='127.0.0.1', port=4001):
        self.etcd = etcd.Client(host=host, port=port)

    def trigger_update(self, new_version):
        self.etcd.write('/current_version', new_version)

    def pending_update(self):
        return not (self.current_version() == self.local_version())

    def current_version(self):
        return self.etcd.read('/current_version').value.strip()

    def ping(self):
        try:
            return bool(self.etcd.machines)
        except etcd.EtcdError, HTTPError:
            return False

    def update_own_info(self, hostname, interval=60, version=None):
        version = version or self.local_version()
        version_dir = '/running_version/%s' % version
        self.etcd.write('%s/%s' % (version_dir, hostname), str(time.time()))
        self.etcd.write(version_dir, None, dir=True, prevExist=True, ttl=(interval*2+10))

    def running_versions(self):
        res = self.etcd.read('/running_version')
        return filter(lambda x: x != 'running_version',
                      [x.key.split('/')[-1] for x in res.children])

    def verify_hosts(self, version, hosts):
        version_dir = '/running_version/%s' % (version,)
        res = self.etcd.read(version_dir)
        if len(list(res.children)) == 1:
            if list(res.children)[0].key == version_dir:
                return []
        hosts_at_version = set([x.key.split('/')[-1] for x in res.children])
        return set(hosts) == hosts_at_version

    def check_single_version(self, version, verbose=False):
        desired_version_seen = False
        running_versions = self.running_versions()
        unwanted_versions = filter(lambda x:x != version,
                                   running_versions)
        wanted_version_found = version in running_versions
        if verbose:
            print 'Wanted version found:', wanted_version_found
            print 'Unwanted versions found:', ', '.join(unwanted_versions)
        return wanted_version_found and not unwanted_versions

    def new_discovery_token(self, discovery_endpoint):
        http = urllib3.PoolManager()
        r = http.request('GET', '%snew' % (discovery_endpoint,))
        return r.data.split('/')[-1]

    def local_version(self, new_value=None):
        mode = new_value is None and 'r' or 'w'

        try:
            with open('/etc/current_version', mode) as fp:
                if new_value is None:
                    return fp.read().strip()
                else:
                    fp.write(new_value)
                    return new_value
        except IOError, e:
            if e.errno == errno.ENOENT:
                return ''
            raise

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Utility for orchestrating updates')
    parser.add_argument('--host', type=str, default='127.0.0.1', help="etcd host")
    parser.add_argument('--port', type=int, default=4001, help="etcd port")
    subparsers = parser.add_subparsers(dest='subcmd')

    trigger_parser = subparsers.add_parser('trigger_update', help='Trigger an update')
    trigger_parser.add_argument('version', type=str, help='Version to deploy')

    current_version_parser = subparsers.add_parser('current_version', help='Get available version')

    ping_parser = subparsers.add_parser('ping', help='Ping etcd')

    pending_update = subparsers.add_parser('pending_update', help='Check for pending update')

    local_version_parser = subparsers.add_parser('local_version', help='Get or set local version')
    local_version_parser.add_argument('version', nargs='?', help="If given, set this as the local version")

    update_own_info_parser = subparsers.add_parser('update_own_info', help="Update host's own info")
    update_own_info_parser.add_argument('--interval', type=int, default=60, help="Update interval")
    update_own_info_parser.add_argument('--hostname', type=str, default=socket.gethostname(),
                                        help="This system's hostname")
    update_own_info_parser.add_argument('--version', type=str,
                                        help="Override version to report into etcd")

    running_versions_parser = subparsers.add_parser('running_versions', help="List currently running versions")

    verify_hosts_parser = subparsers.add_parser('verify_hosts', help="Verify that list of hosts are all available")
    verify_hosts_parser.add_argument('version', help="Version to look for")

    new_discovery_token_parser = subparsers.add_parser('new_discovery_token', help="Get new discovery token")
    new_discovery_token_parser.add_argument('--endpoint', default='https://discovery.etcd.io/', help='Discovery token endpoint')

    check_single_version_parser = subparsers.add_parser('check_single_version', help="Check if the given version is the only one currently running")
    check_single_version_parser.add_argument('version', help='The version to check for')
    check_single_version_parser.add_argument('--verbose', '-v', action='store_true', help='Be verbose')
    args = parser.parse_args()

    do = DeploymentOrchestrator(args.host, args.port)
    if args.subcmd == 'trigger_update':
        do.trigger_update(args.version)
    elif args.subcmd == 'current_version':
        print do.current_version()
    elif args.subcmd == 'check_single_version':
        sys.exit(not do.check_single_version(args.version, args.verbose))
    elif args.subcmd == 'update_own_info':
        do.update_own_info(args.hostname, version=args.version)
    elif args.subcmd == 'ping':
        did_it_work = do.ping()
        if did_it_work:
            print 'Connection succesful'
            sys.exit(0)
        else:
            print 'Connection failed'
            sys.exit(1)
    elif args.subcmd == 'local_version':
        print do.local_version(args.version)
    elif args.subcmd == 'running_versions':
        print '\n'.join(do.running_versions())
    elif args.subcmd == 'new_discovery_token':
        print do.new_discovery_token(args.endpoint)
    elif args.subcmd == 'verify_hosts':
        buffer = sys.stdin.read().strip()
        hosts = buffer.split('\n')
        sys.exit(not do.verify_hosts(args.version, hosts))
    elif args.subcmd == 'pending_update':
        pending_update = do.pending_update()
        if pending_update:
            print 'Yes, there is an update pending'
        else:
            print 'No updates pending'
        sys.exit(not pending_update)



#!/usr/bin/env python
import argparse
import IPy
import os
from novaclient import client as novaclient

"""
Various utils
"""

def get_nova_creds_from_env():
    d = {}
    d['username'] = os.environ['OS_USERNAME']
    d['api_key'] = os.environ['OS_PASSWORD']
    d['auth_url'] = os.environ['OS_AUTH_URL']
    d['project_id'] = os.environ['OS_TENANT_NAME']
    d['region_name'] = os.environ.get('OS_REGION_NAME')
    return d

def get_nova_client():
    return novaclient.Client("1.1", **get_nova_creds_from_env())

def is_rfc1918(ip_string):
    ip = IPy.IP(ip_string)
    if ip in IPy.IP('10.0.0.0/8'):
        return True
    if ip in IPy.IP('192.168.0.0/16'):
        return True
    if ip in IPy.IP('172.16.0.0/12'):
        return True
    return False

def get_ip_of_node(nova_client, name):
    ip = None
    for server in nova_client.servers.list():
        if server.name.split('_')[-1] == name:
            for network in server.networks.values():
                for ip in network:
                    if not is_rfc1918(ip):
                        return ip
            # Fallthrough... If none are non-rfc1918 just return whatever
            return ip
        raise Exception('Server not found')

if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    subparsers = argparser.add_subparsers(dest='action', help='Action to perform')

    get_ip_of_node_parser = subparsers.add_parser('get_ip_of_node', help='Get IP for node')
    get_ip_of_node_parser.add_argument('node_name', help='Node name')

    args = argparser.parse_args()
    nova_client = get_nova_client()

    if args.action == 'get_ip_of_node':
        print get_ip_of_node(nova_client, args.node_name)



#!/usr/bin/env python
import argparse
import mock
import os
import StringIO
import time
import unittest
import yaml
from novaclient import client as novaclient

"""
Parses a specification of nodes to install and makes it so
"""

def get_nova_creds_from_env():
    d = {}
    d['username'] = os.environ['OS_USERNAME']
    d['api_key'] = os.environ['OS_PASSWORD']
    d['auth_url'] = os.environ['OS_AUTH_URL']
    d['project_id'] = os.environ['OS_TENANT_NAME']
    d['region_name'] = os.environ.get('OS_REGION_NAME')
    return d

class ApplyResources(object):
    def __init__(self):
        self.nova_client = None
        self._images = {}
        self._flavors = {}

    def read_resources(self, path):
        fp = file(path)
        return yaml.load(fp)['resources']

    def get_nova_client(self):
        if not self.nova_client:
            self.nova_client = novaclient.Client("1.1", **get_nova_creds_from_env())
        return self.nova_client

    def get_existing_servers(self, project_tag=None, attr_name='name'):
        """
        This method accepts an option project tag
        """
        # NOTE we should check for servers only in a certain state
        nova_client = self.get_nova_client()
        servers = nova_client.servers.list()
        if project_tag:
            servers = [elem for elem in servers if elem.name.endswith('_' + project_tag) ]
        return [getattr(s, attr_name) for s in servers]


    def generate_desired_servers(self, resources, project_tag=None):
        """
        Convert from a hash of servers resources to the
        hash of all server names that should be created
        """
        suffix = (project_tag and ('_' + project_tag)) or ''

        servers_to_create = []
        for k,v in resources.iteritems():
            for i in range(int(v['number'])):
                # NOTE ideally, this would not contain the caridinatlity
                # b/c it is not needed by this hash
                server = {'name': "%s%d%s"%(k, i+1, suffix)}
                servers_to_create.append(dict(server.items() + v.items()))
        return servers_to_create

    def servers_to_create(self, resource_file, project_tag=None):
        resources = self.read_resources(resource_file)
        existing_servers = self.get_existing_servers(project_tag=project_tag)
        desired_servers = self.generate_desired_servers(resources, project_tag)
        return [elem for elem in desired_servers if elem['name'] not in existing_servers ]

    def create_servers(self, servers, userdata, key_name=None):
        for s in servers:
            userdata_file = file(userdata)
            self.create_server(userdata_file, key_name, **s)

    def create_server(self,
                      userdata_file,
                      key_name,
                      name,
                      flavor,
                      image,
                      networks,
                      **keys):
        print "Creating server %s"%(name)
        nova_client = self.get_nova_client()
        self._images[image] = self._images.get(image, nova_client.images.get(image))
        self._flavors[flavor] = self._flavors.get(flavor, nova_client.flavors.get(flavor))
        net_list=[{'net-id': n} for n in networks]
        instance = nova_client.servers.create(
          name=name,
          image=self._images[image],
          flavor=self._flavors[flavor],
          nics=net_list,
          userdata=userdata_file,
          key_name=key_name,
        )

        # Poll at 5 second intervals, until the status is no longer 'BUILD'
        status = instance.status
        while status == 'BUILD':
            time.sleep(5)
            # Retrieve the instance again so the status field updates
            instance = nova_client.servers.get(instance.id)
            status = instance.status
            print "status: %s" % status

    def delete_servers(self, project_tag):
        nova_client = self.get_nova_client()
        servers = self.get_existing_servers(project_tag=project_tag, attr_name='id')
        for uuid in servers:
            print "Deleting uuid: %s"%(uuid)
            nova_client.servers.delete(uuid)

class TestApplyResources(unittest.TestCase):
    server_data = [('foo1_abc123', '93138146-2275-4e18-b41e-3957aa13e73a'),
                   ('foo2_abc124', '26af0276-83e1-4b68-870e-ff3250be8e8f'),
                   ('foo4_bc124', '677388b7-b5ac-418b-b671-6b930dc8003a'),
                   ('bar2', '381877b2-12c5-4831-95ed-1d7518bb7e8c'),
                   ('baz', '59e5dd8d-2063-4943-98de-df206e462849')]

    def fake_server_data(self, nova_client):
        def fake_server(name, uuid):
            s = mock.Mock()
            s.configure_mock(name=name, id=uuid)
            return s
        server_list = [fake_server(*s) for s in self.server_data]
        nova_client.servers.list.return_value = server_list

    def test_get_existing_servers(self):
        apply_resources = ApplyResources()
        with mock.patch.object(apply_resources, 'get_nova_client') as get_nova_client:
            nova_client = get_nova_client.return_value
            self.fake_server_data(nova_client)
            self.assertEquals(apply_resources.get_existing_servers(), [s[0] for s in self.server_data])
            self.assertEquals(apply_resources.get_existing_servers(project_tag='abc123'),
                              ['foo1_abc123'])
            self.assertEquals(apply_resources.get_existing_servers(project_tag='abc124'),
                              ['foo2_abc124'])
            self.assertEquals(apply_resources.get_existing_servers(project_tag='bc124'),
                              ['foo4_bc124'])
            self.assertEquals(apply_resources.get_existing_servers(project_tag='bc124', attr_name='id'),
                              ['677388b7-b5ac-418b-b671-6b930dc8003a'])

    def test_generate_desired_servers(self):
        apply_resources = ApplyResources()
        self.assertEquals(apply_resources.generate_desired_servers({'foo': {'number': 5 }}, project_tag='foo'),
                          [{'name': 'foo1_foo', 'number': 5},
                           {'name': 'foo2_foo', 'number': 5},
                           {'name': 'foo3_foo', 'number': 5},
                           {'name': 'foo4_foo', 'number': 5},
                           {'name': 'foo5_foo', 'number': 5}
                          ])
        self.assertEquals(apply_resources.generate_desired_servers({'foo': {'number': 5 },
                                                                    'bar': {'number': 2 }}, project_tag='foo'),
                          [{'name': 'foo1_foo', 'number': 5},
                           {'name': 'foo2_foo', 'number': 5},
                           {'name': 'foo3_foo', 'number': 5},
                           {'name': 'foo4_foo', 'number': 5},
                           {'name': 'foo5_foo', 'number': 5},
                           {'name': 'bar1_foo', 'number': 2},
                           {'name': 'bar2_foo', 'number': 2},
                          ])
        self.assertEquals(apply_resources.generate_desired_servers({'foo': {'number': 0 },
                                                                    'bar': {'number': 2 }}),
                          [{'name': 'bar1', 'number': 2},
                           {'name': 'bar2', 'number': 2}])

    def test_servers_to_create(self):
        apply_resources = ApplyResources()
        with mock.patch.multiple(apply_resources,
                                 get_nova_client=mock.DEFAULT,
                                 read_resources=mock.DEFAULT) as mocks:
            get_nova_client = mocks['get_nova_client']
            read_resources = mocks['read_resources']
            read_resources.return_value = {'foo': {'number': 1},
                                           'bar': {'number': 2}}

            nova_client = get_nova_client.return_value
            self.fake_server_data(nova_client)

            self.assertEquals(apply_resources.servers_to_create('fake_path'),
                              [{'name': 'foo1', 'number': 1},
                               {'name': 'bar1', 'number': 2}])

    def test_create_servers(self):
        apply_resources = ApplyResources()
        with mock.patch('__builtin__.file') as file_mock:
            with mock.patch.object(apply_resources, 'create_server') as create_server:
                file_mock.side_effect = lambda f: StringIO.StringIO('test user data')
                apply_resources.create_servers([{'name': 'foo1', 'networks':  ['someid']},
                                                {'name': 'foo2', 'networks':  ['someid']}], 'somefile', 'somekey')
                create_server.assert_any_call(mock.ANY, 'somekey', name='foo1', networks=['someid'])
                create_server.assert_any_call(mock.ANY, 'somekey', name='foo2', networks=['someid'])
                for call in create_server.call_args_list:
                    self.assertEquals(call[0][0].read(), 'test user data')

if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    subparsers = argparser.add_subparsers(dest='action', help='Action to perform')

    apply_parser  = subparsers.add_parser('apply', help='Apply a resource file')
    apply_parser.add_argument('resource_file_path', help='Path to resource file')
    apply_parser.add_argument('userdata', help='Path of userdata to apply to all nodes')
    apply_parser.add_argument('--project_tag', help='Project tag')
    apply_parser.add_argument('--key_name', help='Name of key pair')

    delete_parser = subparsers.add_parser('delete', help='Delete a project')
    delete_parser.add_argument('project_tag', help='Id of project to delete')

    list_parser  = subparsers.add_parser('list', help='List servers described in resource file')
    list_parser.add_argument('resource_file_path', help='Path to resource file')
    list_parser.add_argument('--project_tag', help='Project tag')

    args = argparser.parse_args()
    if args.action == 'apply':
        apply_resources = ApplyResources()
        servers = apply_resources.servers_to_create(args.resource_file_path,
                                                     project_tag=args.project_tag)
        apply_resources.create_servers(servers, args.userdata, key_name=args.key_name)
    elif args.action == 'delete':
        if not args.project_tag:
            argparser.error("Must set project tag when action is delete")
        ApplyResources().delete_servers(project_tag=args.project_tag)
    elif args.action == 'list':
        apply_resources = ApplyResources()
        resources = apply_resources.read_resources(args.resource_file_path)
        desired_servers = apply_resources.generate_desired_servers(resources, args.project_tag)
        print '\n'.join([s['name'] for s in desired_servers])
